/* 
******************************************************************************

   S3 86c375 (ViRGE) integrated RAMDAC driver

   Copyright (C) 1998, 1999 Jos Hulzink	[josh@stack.nl]

   Permission is hereby granted, free of charge, to any person obtaining a
   copy of this software and associated documentation files (the "Software"),
   to deal in the Software without restriction, including without limitation
   the rights to use, copy, modify, merge, publish, distribute, sublicense,
   and/or sell copies of the Software, and to permit persons to whom the
   Software is furnished to do so, subject to the following conditions:

   The above copyright notice and this permission notice shall be included in
   all copies or substantial portions of the Software.

   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
   IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
   THE AUTHOR(S) BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
   IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
   CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

******************************************************************************
*/

#include <kgi/maintainers.h>
#define	MAINTAINER	"Jos_Hulzink (josh@stack.nl)"
#define	DRIVER_NAME	"S3 ViRGE integrated RAMDAC"
#define	DRIVER_REV	"$Revision: 1.13 $"

#define	DEBUG_LEVEL	0
#define DAC_PRECISION	6
#define STREAMS_FOR_8BPP

#include <kgi/module.h>
#include <kgi/types.h>
#include <kgi/kgi_commands.h>
#include <S3/sdac.h>
#include <S3/virge.h>

#define	MODES (GM_TEXT16|GM_1BIT|GM_4BIT|GM_8BIT|GM_15BIT|GM_16BIT|GM_32BIT)

static struct kgi_ramdac ramdac = {
	"S3 Incorporated",			/* Manufacturer		*/
	"ViRGE",				/* Model		*/
	GM_TEXT16 | GM_1BIT | GM_4BIT |		/* Supported modes	*/
	GM_8BIT | GM_15BIT | GM_16BIT |
	GM_24BIT | GM_32BIT,
	FL_DAC_8BIT,				/* Flags		*/
	{ 2048, 2048 },				/* Maximum resolution	*/
	NULL,					/* Private data hook	*/
	DAC_8BIT_PB,				/* Device type		*/
	{ 0,  135000000 },			/* Normal mode		*/
	{ 0,  135000000 },			/* 2xclk mode		*/
	{
		{ 0,  75000000 },		/* KGIGT_TEXT		*/
		{ 0,         0 },		/* KGIGT_TEXT32		*/
		{ 0,  94500000 },		/* KGIGT_1BIT		*/
		{ 0,  94500000 },		/* KGIGT_4BIT		*/
		{ 0, 135000000 },		/* KGIGT_8BIT		*/	
		{ 0,  94500000 },		/* KGIGT_15BIT		*/
		{ 0,  94500000 },		/* KGIGT_16BIT		*/
		{ 0,  56600000 },		/* KGIGT_24BIT		*/
		{ 0,  56600000 }		/* KGIGT_32BIT		*/
	}
};

/* Driver state info 
*/
struct dac_private 
{
  kgi_uint	cur_hotx, cur_hoty;		/* Cursor hot spot	*/
  kgi_uint	cur_x, cur_y;			/* Current position	*/
  kgiu8	ctrl0, ctrl1, ctrl2, ctrl3;	        /* Control registers	*/
};

struct virge_streams_struct
{
  kgiu32 offset;
  kgiu32 data;
} virge_streams[] = 
{
  { 0x0180, 0x00000000 } , /*  0 Primary Stream Control	*/
  { 0x0184, 0x00000000 } , /*  1 Color/Chroma Key Control*/
  { 0x0190, 0x03000000 } , /*  2 Secondary Stream Control*/
  { 0x0194, 0x00000000 } , /*  3 Chroma Key Upper Bound */
  { 0x0198, 0x00000000 } , /*  4 Secondary Stream Stretch Factor */
  { 0x01A0, 0x01000000 } , /*  5 Blend Control */
  { 0x01C0, 0x00000000 } , /*  6 Primary Stream Frame Buffer Address 0 */
  { 0x01C4, 0x00000000 } , /*  7 Primary Stream Frame Buffer Address 1 */
  { 0x01C8, 0x00000780 } , /*  8 Primary Stream Stride */
  { 0x01CC, 0x00000000 } , /*  9 Double Buffer support */
  { 0x01D0, 0x00000000 } , /* 10 Secondary Stream Frame Buffer Address 0 */
  { 0x01D4, 0x00000000 } , /* 11 Secondary Stream Frame Buffer Address 1 */
  { 0x01D8, 0x00000001 } , /* 12 Secondary Stream Stride */
  { 0x01DC, 0x48000008 } , /* 13 Opaque overlay control */
  { 0x01E0, 0x00000000 } , /* 14 K1 vertical scale factor*/
  { 0x01E4, 0x00000000 } , /* 15 K2 vertical scale factor*/
  { 0x01E8, 0x00000000 } , /* 16 Vertical Accumulator Initial Value */
  { 0x01EC, 0x00000000 } , /* 17 FIFO Control */
  { 0x01F0, 0x00000001 } , /* 18 Primary stream window base (0,0) */
  { 0x01F4, 0x027F01E0 } , /* 19 Primary stream window size (width,height) */
  { 0x01F8, 0x00000000 } , /* 20 Secondary stream window base */
  { 0x01FC, 0x00020001 } , /* 21 Secondary stream window size */
  { 0x0200, 0x00006000 } , /* 22 FIFO Control */
}; 

#define	CTRL0		private->ctrl0
#define	CTRL1		private->ctrl1
#define	CTRL2		private->ctrl2
#define	CTRL3		private->ctrl3
#define	CUR_HOTX	private->cur_hotx
#define	CUR_HOTY	private->cur_hoty
#define	CUR_X		private->cur_x
#define	CUR_Y		private->cur_y
#undef  GET_PRIVATE    
#define	GET_PRIVATE	struct dac_private *private = KGIM_PRIVATE(dpy,ramdac)
#define MEM_VGA		(((struct chipset_private*)(KGIM_PRIVATE(dpy,chipset)))->mem_vga)

#include <IBM/clut.inc>

extern void CRT_out8 (struct kgi_display *,kgiu8,kgiu32);
extern kgiu8 CRT_in8 (struct kgi_display *,kgiu32);
extern void SEQ_out8 (struct kgi_display *,kgiu8,kgiu32);
extern kgiu8 SEQ_in8 (struct kgi_display *,kgiu32);
extern void SPR_out32(struct kgi_display *,kgiu32, kgiu32);

/* The streams processor has to be enabled for all 24/32 bpp modes
For 15/16 bpp, enabling streams fixes the pixelshift problem. The pixelshift
bug is still in the chipset driver. 
For 8 bpp, streams shouldn't be needed at all, but in some modes it is nice
(now it is the only way to get lowres modes working, and some day streams will
handle hardware doublebuffering support, so enable it if you can)
*/

/* Do NOT enable streams for : interlaced, 1/4 bpp, or CLK doubling modes */

void disable_streams_processor (struct kgi_display *dpy)
{
  kgiu8 i;
  if ((CRT_in8 (dpy, 0x67) & 0x0C)!=0x0C) return;
  /* Find VSYNC */
  for (i=0; i < 10000000; i++) 
    if (mem_in8(MEM_VGA + 0x03DA) & 0x08) break;
  SPR_out32(dpy,0x0200,0x0000C000);
  CRT_out8 (dpy, CRT_in8 (dpy, 0x67) & ~0x0C, 0x67); 
}

void set_streams_processor (struct kgi_display *dpy, struct kgi_mode *mode )
{
  IS_CRITICAL;
  register kgiu8 i;
    
  DEBUG1("Preparing Streams processor for %dx%d#%dx%d mode, graphtype %d",
  	  mode->request.visible.x,
  	  mode->request.visible.y,
  	  mode->request.virt.x,
  	  mode->request.virt.y,
  	  mode->request.graphtype);
  switch (mode->request.graphtype) {
    case KGIGT_8BIT:
      virge_streams[0].data = 0x00000000;
      virge_streams[8].data = mode->request.virt.x;
      break;
    case KGIGT_15BIT:
      virge_streams[0].data = 0x03000000;
      virge_streams[8].data = 2*(mode->request.virt.x);
      break;
    case KGIGT_16BIT:
      virge_streams[0].data = 0x05000000;
      virge_streams[8].data = 2*(mode->request.virt.x);
      break;
    case KGIGT_24BIT:
      virge_streams[0].data = 0x06000000;
      virge_streams[8].data = 3*(mode->request.virt.x);
      virge_streams[22].data= 0xc000;
      break;
    case KGIGT_32BIT:
      virge_streams[0].data = 0x07000000;
      virge_streams[8].data = 4*(mode->request.virt.x);
      virge_streams[22].data= 0x10000;
      break;
    default:
      DEBUG1("Wrong colour depth for Streams.");
      return;
  }
  virge_streams[02].data= mode->request.visible.x-1;
  virge_streams[14].data= 
  virge_streams[16].data= mode->y.width-1;
  virge_streams[19].data= ((mode->x.width-1) << 16) + 
  			    mode->y.width; 
  /* Switch to pixeldoubling mode if requested screen is lowres */
  if (mode->magnify.x==2) {
    /* Primary stream 2x stretch */
    virge_streams[0].data |= 0x20000000;
    virge_streams[02].data= -1 & 0x0fff;
    virge_streams[04].data= ((-mode->request.visible.x) & 0x7ff) * 0x10001;
  } 
    /* Double vertical line count */
  if (mode->magnify.y==2) {
    virge_streams[15].data = (-(mode->request.visible.y)) & 0x000007FF;
  } else {
    virge_streams[15].data = 0; /* Vertical output equals frame size. */
  }
  ENTER_CRITICAL; 
  for (i=0;i<23;i++)
    SPR_out32(dpy,virge_streams[i].offset,virge_streams[i].data);
  LEAVE_CRITICAL;
}

/* KGI interface 
*/
int kgim_ramdac_init(struct kgi_card *card, struct kgi_display *dpy)
{
	
	struct dac_private *private;
	if (!(private = kgi_malloc_small(sizeof(*private)))) 
	{
		
		NOTICE1("failed to allocate private data!");
		return -ENOMEM;
	}
	
	KGIM_PRIVATE(dpy, ramdac) = private;
	memset(private, 0, sizeof(*private));
	card->ramdac = &ramdac;
	
	return EOK;
}

void kgim_ramdac_done(struct kgi_display *dpy)
{
	disable_streams_processor (dpy);
	kgi_free_small(KGIM_PRIVATE(dpy, ramdac));
	KGIM_PRIVATE(dpy, ramdac) = NULL;
}

int kgim_ramdac_check_mode (struct kgi_display *dpy, struct kgi_mode *mode, 
			    enum kgi_tm_cmd *cmd)
{
	int div;
	int newclock = mode->dclk;
	struct kgi_ramdac *dac = dpy->card->ramdac;
	
	switch(mode->request.graphtype) 
	{
		
	case KGIGT_TEXT16:	
		div = 1;
		break;
		
	case KGIGT_1BIT:
		div = 8;
		break;
		
	case KGIGT_4BIT:
		div = 8;
		break;
		
	case KGIGT_8BIT:
		div = 4;
		break;
		
	case KGIGT_15BIT:
	case KGIGT_16BIT:	
		div = 2;
		break;
		
	case KGIGT_24BIT:
	case KGIGT_32BIT:	
		div = 1;
		break;
		
	default:
		DEBUG1("Could not handle mode %x", mode->request.graphtype);
		return -E(RAMDAC, UNKNOWN);
	}

	switch (*cmd) 
	{
		
	case CMD_PROPOSE:
		mode->lclk.mul = 1;
		mode->clk.mul = 1; 
		mode->clk.div = 1;
		mode->lclk.div = div;
		DEBUG1("%i:%i multiplex mode", mode->lclk.div, mode->lclk.mul);
		mode->flags &= ~FL_DAC_2XCLOCK;
		mode->flags |= FL_DAC_8BIT;
		if ((mode->request.graphtype == KGIGT_32BIT) ||
		    (mode->request.graphtype == KGIGT_24BIT) || 
		    (mode->request.graphtype == KGIGT_15BIT) ||
		    (mode->request.graphtype == KGIGT_16BIT)) 
			mode->flags |= FL_DAC_GAMMA;
			
		return EOK;
		
	case CMD_LOWER:
		if ((newclock < dac->clock[mode->request.graphtype].min) ||
		    (newclock < dac->normal.min)) 
		{
			DEBUG1("%i Hz DCLK is too low", newclock);
			return -E(RAMDAC, UNKNOWN);
		}
		
		if (dac->clock[mode->request.graphtype].max < newclock) 
		  newclock = dac->clock[mode->request.graphtype].max;
		
		if (dac->normal.max < newclock) 
		  if ((mode->request.graphtype == KGIGT_8BIT) &&
		      (newclock > dac->dblclk.max)) 
		    newclock = dac->dblclk.max;
		
		DEBUG1("DCLK = %i", newclock);
		mode->dclk = newclock;
		return EOK;
		
	case CMD_RAISE:
		if ((newclock > dac->clock[mode->request.graphtype].max) ||
		    (newclock > (mode->request.graphtype == KGIGT_8BIT?
		    		 dac->dblclk.max:dac->normal.max))) {
		  DEBUG1("%i Hz DCLK is too high", newclock);
		  return -E(RAMDAC, UNKNOWN);
		}
		
		if (newclock < dac->clock[mode->request.graphtype].min) 
  		  newclock = dac->clock[mode->request.graphtype].min;
		
		if (newclock < dac->normal.min) 
  		  newclock = dac->normal.min;
		
		mode->dclk = newclock;
		return EOK;
		
	case CMD_CHECK:
		/* Check multiply parameters */
		if ((mode->lclk.mul != 1) || (mode->lclk.div != div) ||
		    (mode->clk.mul != 1) || (mode->clk.div != 1)) 
		  return -E(RAMDAC, UNKNOWN);
		/* Check for lowest limit. */
		if ((mode->dclk < dac->normal.min) ||
		    (mode->dclk < dac->clock[mode->request.graphtype].min))
		  return -E(RAMDAC, UNKNOWN);
		/* Check for highest limit. */
		if (mode->dclk > dac->normal.max) { 
	  	  if (mode->request.graphtype == KGIGT_8BIT) { 
	    	    if (mode->dclk > dac->dblclk.max)
	  	      return -E(RAMDAC, UNKNOWN);
	  	  }
	  	  else return -E(RAMDAC, UNKNOWN);
	  	}
		if (mode->dclk > dac->clock[mode->request.graphtype].max)
		  return -E(RAMDAC, UNKNOWN);
		  
		return EOK;
		
	default:
		INTERNAL_ERROR;
		return -E(RAMDAC, UNKNOWN);
	}
}

void kgim_ramdac_set_mode(struct kgi_display *dpy,struct kgi_mode *mode)
{
/* Changed registers:
** CR67, SR08, SR15, SR18, All Streams MMIO registers
*/
        kgiu8 cr67=0,sr15,sr18,t;
	struct kgi_ramdac *dac = dpy->card->ramdac;

	SEQ_out8(dpy, 0x06, 0x08);
	sr15 = SEQ_in8(dpy, 0x15) & ~0x50;
	sr18 = 0;
	
	if (mode->dclk >= dac->normal.max)
	  if (mode->request.graphtype != KGIGT_8BIT) 
	    INTERNAL_ERROR;
  
  	disable_streams_processor (dpy);
	switch (mode->request.graphtype) 
	{
		
	case KGIGT_1BIT:
	case KGIGT_TEXT16:
	case KGIGT_4BIT:
	  cr67=CR67_DAC_MODE0;
	  break;
	case KGIGT_8BIT:
	  if (mode->dclk >= dac->normal.max) {
	    cr67=CR67_DAC_MODE8 | CR67_VCLK_PHS;
	    #ifdef STREAMS_FOR_8BPP
	    cr67 |= 0x0C;
	    #endif 
 	    sr15 |= 0x50;
	    sr18 = 0x80;
	  }
	  else {
	    cr67=CR67_DAC_MODE0;
	    #ifdef STREAMS_FOR_8BPP
	    cr67 |= 0x0C;
	    #endif 
	  } 
	  break;
	case KGIGT_15BIT:
	  cr67=CR67_DAC_MODE9 | 0x0C; 
	  break;
	case KGIGT_16BIT:
	  cr67=CR67_DAC_MODEA | 0x0C; 
	  break;
	case KGIGT_24BIT:
	  cr67=CR67_DAC_MODED | 0x0C;
	  break;
	case KGIGT_32BIT:
	  cr67=CR67_DAC_MODED | 0x0C;
	  break;
	default:
 	  INTERNAL_ERROR;
	}
	
	SEQ_out8(dpy, sr15, 0x15);
	SEQ_out8(dpy, sr18, 0x18);
	/* Wait for vertical retrace before setting CRT67 */
	for (t=0; t < 10000000; t++) 
	  if (mem_in8(MEM_VGA + 0x03DA) & 0x08) break;
	CRT_out8(dpy, cr67, 0x67);
	set_streams_processor (dpy,mode); 
	
	dpy->set_clut = dac_set_clut;
}

int kgim_ramdac_command(struct kgi_graphic *graph, kgi_uint cmd, kgi_uint arg) 
{
	/* This may set:
	 * - Hardware cursor
	 * - CLUT (or gamma correction table)
	 */
	kgi_clut clut;

	switch (cmd) 
	{ 
	case RAMDAC_GETINFO:
		graph->io_res = &ramdac;
		return EOK;
		
	case RAMDAC_SETINFO:
		if (!suser())
		  return -EPERM; 
		memcpy(&ramdac, graph->io_buf, sizeof(ramdac));
		return EOK;
		
	case RAMDAC_SETCLUT:
		clut.size = 256;
		clut.data = (kgi_color *)(graph->io_buf);
		dac_set_clut(graph->dev.dpy, &clut, 0, 256);
		return EOK;
		
	case RAMDAC_GETCLUT:
		clut.size = 256;
		clut.data = (kgi_color *)(graph->io_buf);
		graph->io_res = graph->io_buf;
		dac_get_clut(graph->dev.dpy, &clut, 0, 256);
		return EOK;
	default:
	}
	return -E(RAMDAC,NOSUP_ALWAYS_CANT);  /* The driver can't do this */
}
